#
# (c) 2021-2022 Giorgio Gonnella, University of Goettingen, Germany
#
"""
Programmatically import a Python module and compile it on the
fly if it is an extension module written in Nim or Rust.
"""

from pathlib import Path
from multiplug.py_mod import py
from multiplug.rust_mod import rust, RUST_CONST_CLS
from multiplug.nim_mod import nim, NIM_CONST_PFX
from multiplug.bash_mod import bash
from multiplug.error import UnsupportedLangError

def importer(filename, verbose=False, req_const=[], opt_const=[],
             req_func=[], opt_func=[], rust_const_cls=RUST_CONST_CLS,
             nim_const_pfx=NIM_CONST_PFX, disable_bash=False):
  """
  Import a module given its "filename".

  The module can be:
  - a pure Python module
  - an extension module written in Nim or Rust
    (which is automatically compiled if necessary)
  - a Bash script
    (for which a wrapper module is automatically generated)

  The language of the module is deducted from the filename extension, which
  must be one of: ".py", ".nim", ".rs", ".sh".
  If ``disable_bash`` is set, then Bash scripts are not supported.

  Return value:
    the imported module

  Attributes added to the imported module:
    __lang__: the language of the module (one of "py", "rust", "nim", "bash")
    variables/constants:
      Python: as in normal module import
      Nim: constants passed to exportpy_consts or exportpy_consts_wpfx;
           functions exported to Python with name starting with ``py_const_``
           (or the value of ``nim_const_pfx``)
      Rust: constants defined in a Struct, exported to Python,
            named ``Constants`` (or the value of ``rust_const_cls``)
      Bash: variables whose name does not start with _
            (with the value they have after script execution)
    functions:
      Python: as in normal module import
      Nim: functions exported to Python --
           except those with name starting with ``py_const_``
           (or the value of ``nim_const_pfx``)
      Rust: functions exported to Python
      Bash: functions whose name does not start with _,
            wrapped in a function with parameters ``(*args, **kwargs)``
  """
  suffix = Path(filename).suffix
  if suffix == ".rs":
    m = rust(filename, verbose=verbose, req_const=req_const,
             opt_const=opt_const, req_func=req_func, opt_func=opt_func,
             const_cls=rust_const_cls)
  elif suffix == ".sh":
    if disable_bash:
      raise UnsupportedLangError("Bash scripts are not supported by this "+\
                                  "application")
    m = bash(filename, verbose=verbose, req_const=req_const,
             opt_const=opt_const, req_func=req_func, opt_func=opt_func)
  elif suffix == ".nim":
    m = nim(filename, verbose=verbose, req_const=req_const,
            opt_const=opt_const, req_func=req_func, opt_func=opt_func,
            const_pfx=nim_const_pfx)
  elif suffix == ".py":
    m = py(filename, verbose=verbose, req_const=req_const,
           opt_const=opt_const, req_func=req_func, opt_func=opt_func)
  else:
    raise UnsupportedLangError("Unsupported language, file suffix is: " \
                                 + suffix)
  return m



#
# (c) 2021-2022 Giorgio Gonnella, University of Goettingen, Germany
#
##
## Check the API of a plugin module
##

from multiplug.error import InterfaceRequirementError

def _enforce_section(m, mname, units, unit_type, required):
  found_units = {True: [], False: []}
  for u in units:
    found = hasattr(m, u)
    if not found:
      if required:
        raise InterfaceRequirementError(
            "Plugin module '{}' does not define the required {} '{}'".format(
                mname, unit_type, u))
      else:
        setattr(m, u, None)
    found_units[found].append(u)
  info = []
  lbl = "required" if required else "optional"
  if found_units[True]:
    info.append("#   imported {} {}: {}\n".format(lbl, unit_type,
                                                ", ".join(found_units[True])))
  if found_units[False]:
    info.append("#   non defined {} {} (set to None): {}\n".\
        format(lbl, unit_type, ", ".join(found_units[False])))
  return info

def enforce_plugin_api(m, mname, req_const=[], req_func=[],
                                 opt_const=[], opt_func=[]):
  """
  - checks if required functions and constants are defined in the module
  - if optional functions and constants are not defined, they are set to None
  - messages showing the list of found and not found
    functions and constants are collected in the returned list
  """
  info = _enforce_section(m, mname, req_const, "constant", True)
  info += _enforce_section(m, mname, opt_const, "constant", False)
  info += _enforce_section(m, mname, req_func, "function", True)
  info += _enforce_section(m, mname, opt_func, "function", False)
  return info




#
# (c) 2021-2022 Giorgio Gonnella, University of Goettingen, Germany
#
import sh
import importlib
from multiplug.error import InterfaceRequirementError
from pathlib import Path
import sys
from multiplug.plugin_api import enforce_plugin_api

def _is_array(filename, cname):
  try:
    sh.bash("-c", f". {filename}; [[ \"$(declare -p {cname})\" "+\
            "=~ \"declare -a\" ]]")
    return True
  except sh.ErrorReturnCode_1:
    return False

def _get_constant(filename, cname):
  is_array = _is_array(filename, cname)
  if is_array:
    value = sh.bash("-c", f". {filename}; " +\
       f"for elem in \"${{{cname}[@]}}\"; do echo -e $elem; done").\
            rstrip().split("\n")
    has_tabs = any("\t" in v for v in value)
    if has_tabs:
      value = [v.split("\t") for v in value]
  else:
    value = sh.bash("-c", f". {filename}; echo -e ${cname}").rstrip()
  return value

def _list_attributes(m, filename, verbose):
  before = sh.mktemp().rstrip()
  after = sh.mktemp().rstrip()
  sh.bash("-c", f"set > {before}")
  sh.bash("-c", f". {filename}; set > {after}")
  varnames=set(sh.bash("-c", f"diff {before} {after} |"+\
      "grep '^>' | cut -c3- | grep '^[A-Za-z]' | grep '=' | "+\
      "grep -o '^[^=]*'; true").rstrip().split("\n"))
  funcnames=set(sh.bash("-c", f"diff {before} {after} |"+\
      "grep '^>' | cut -c3- | grep '^[A-Za-z]' | grep -v '=' | grep '()' |"+\
      "grep -o '^[A-Za-z][A-Za-z0-9_]*'; true").rstrip().split("\n"))
  varnames -= {"BASH_REMATCH", "COMP_WORDBREAKS", "BASH_EXECUTION_STRING",
      "PIPESTATUS"}
  sh.rm(before)
  sh.rm(after)
  return varnames, funcnames


def _wrap_function(filename, funcname):
  def _fn(*args, **kwargs):
    args_str=" ".join(args)
    kwargs_str=" ".join([f"{k}={v}" for k, v in kwargs.items()])
    retval = str(sh.bash("-c",
      f". {filename}; {funcname} {args_str} {kwargs_str}", _err=sys.stderr))
    if len(retval) > 0 and retval[-1] == "\n":
      retval = retval[:-1]
    has_newlines = "\n" in retval
    has_tabs = "\t" in retval
    if has_newlines and has_tabs:
      retval = [l.split("\t") for l in retval.split("\n")]
    elif has_newlines:
      retval = retval.split("\n")
    elif has_tabs:
      retval = retval.split("\t")
    return retval
  return _fn

def bash(filename, verbose=False, req_const=[], opt_const=[],
                                  req_func=[],  opt_func=[]):
  """
  Creates a Python module which wraps a shell script.
  """
  modulename=Path(filename).stem
  spec=importlib.machinery.ModuleSpec(modulename, None)
  m = importlib.util.module_from_spec(spec)

  varnames, funcnames = _list_attributes(m, filename, verbose)
  for v in varnames:
    setattr(m, v, _get_constant(filename, v))
  for f in funcnames:
    setattr(m, f, _wrap_function(filename, f))
  info = [f"# bash module {modulename} imported from file {filename}\n"]
  info += enforce_plugin_api(m, modulename, req_const=req_const,
                           opt_const=opt_const, req_func=req_func,
                           opt_func=opt_func)
  if verbose:
    sys.stderr.write("".join(info))
  m.__lang__ = "bash"
  return m



#
# (c) 2021-2022 Giorgio Gonnella, University of Goettingen, Germany
#
"""
Programmatically import a Python module
"""

import sys
import importlib
from pathlib import Path
from multiplug.plugin_api import enforce_plugin_api

def py(filename, verbose=False, req_const=[], opt_const=[],
                                req_func=[],  opt_func=[]):
  """
  Programmatically import Python module "filename"
  """
  modulename = Path(filename).stem
  spec = importlib.util.spec_from_file_location(modulename, filename)
  m = importlib.util.module_from_spec(spec)
  spec.loader.exec_module(m)

  info = [f"# python module {modulename} imported from file {filename}\n"]
  info += enforce_plugin_api(m, modulename, req_const=req_const,
                           opt_const=opt_const, req_func=req_func,
                           opt_func=opt_func)
  if verbose:
    sys.stderr.write("".join(info))
  m.__lang__ = "python"
  return m




#
# (c) 2022 Giorgio Gonnella, University of Goettingen, Germany
#
"""
Exceptions for the library
"""

class MultiplugError(Exception):
  """parent class for package-specific errors"""
  pass

class InterfaceRequirementError(MultiplugError):
  """raised when a requirement in api_config is not met"""
  pass

class UnsupportedLangError(MultiplugError):
  """raised when a plugin is written in an unsupported language"""
  pass



#
# (c) 2021-2022 Giorgio Gonnella, University of Goettingen, Germany
#
"""
Programmatically import a Python module and compile it on the
fly if it is an extension module written in Nim or Rust.
"""

import sh
import sys
import importlib
from pathlib import Path
import os
from contextlib import contextmanager
from multiplug.plugin_api import enforce_plugin_api

RUST_CONST_CLS = "Constants"

_toml_content = """
[package]
authors = ["*"]
name = "{modulename}"
version = "1.0.0"
edition = "2018"

[dependencies]
pyo3 = {{ version = "0.13.2", features = ["abi3-py36", "extension-module"] }}
{more_dependencies}

[lib]
name = "{modulename}"
crate-type = ["cdylib"]
path = "{modulename}.rs"
"""

def _create_cargo_toml(filename):
  """
  Create Cargo.toml for a pyo3/maturin module in the parent directory of
  <filename>.

  If the Rust code needs any crate, you can add a comment like:
  // [dependencies] foobar = "1.2.3"
  and this is added automatically to the [dependencies] section of Cargo.toml
  """
  workingdir = Path(filename).parent
  modulename = Path(filename).stem
  try:
    more_dependencies=sh.grep('\/\/ \[dependencies\] .*',
                              filename, P=True)
    more_dependencies=str(more_dependencies)[18:]
  except sh.ErrorReturnCode_1:
    more_dependencies=""
  with open(workingdir/"Cargo.toml", "w") as f:
    f.write(_toml_content.format(modulename = modulename,
                                more_dependencies=more_dependencies))

@contextmanager
def _cargo_for(filename):
  """
  Makes sure that a Cargo.toml exists for a module.
  If not, creates one using create_cargo_toml().
  If created, it is also deleted (and the Cargo.lock file too).
  """
  workingdir = Path(filename).parent
  cargo_temporary = False
  if not os.path.exists(workingdir/"Cargo.toml"):
    cargo_temporary = True
    _create_cargo_toml(filename)
  try:
    yield workingdir/"Cargo.toml"
  finally:
    if cargo_temporary:
      sh.rm(workingdir/"Cargo.toml", workingdir/"Cargo.lock")

def _import_constants(m, membername):
  """
  Transfer attributes of a member of a module to the module
  and delete the member.

  Attribute starting with "_" are not transfered.

  If the member does not exist, nothing is done, silently.
  """
  if hasattr(m, membername):
    member = getattr(m, membername)
    for k in list(member.__dict__.keys()):
      if not k.startswith("_"):
        setattr(m, k, getattr(member, k))
    delattr(m, membername)

def rust(filename, verbose=False, req_const=[], opt_const=[],
         req_func=[],  opt_func=[], const_cls=RUST_CONST_CLS):
  """
  Import a Rust module written using pyo3, compiling it with maturin.

  If no Cargo.toml is present (in the parent directory of <filename>),
  one is created by automatically (see _create_cargo_toml
  for details). In this case Cargo.toml and Cargo.lock are
  deleted after running the method.

  The "target" subdirectory (in the parent directory of <filename>) is not
  removed, so that unnecessary recompilations are avoided. You may want to add
  "target" to .gitignore.
  """
  workingdir = Path(filename).parent
  modulename = Path(filename).stem
  with _cargo_for(filename) as cargo:
    sh.maturin("build", release=True, m=cargo)
  libfilename = workingdir/f"{modulename}.so"
  sh.ln(f"target/release/lib{modulename}.so", libfilename, s=True, f=True)
  spec = importlib.util.spec_from_file_location(modulename, libfilename)
  m = importlib.util.module_from_spec(spec)
  spec.loader.exec_module(m)
  info = [f"# rust module {modulename} imported from file {filename}\n"]
  if len(const_cls) > 0:
    _import_constants(m, const_cls)
    info += enforce_plugin_api(m, modulename, req_func=req_func,
                             opt_func=opt_func, req_const=req_const,
                             opt_const=opt_const)
  else:
    assert len(req_const) == 0
    info += enforce_plugin_api(m, modulename, req_func=req_func,
                             opt_func=opt_func, opt_const=opt_const)
    info.append("# constants definition mechanism disabled\n")
  if verbose:
    sys.stderr.write("".join(info))
  m.__lang__ = "rust"
  return m



#
# (c) 2022 Giorgio Gonnella, University of Goettingen, Germany
#
from multiplug.importer import importer
from multiplug.bash_mod import bash
from multiplug.nim_mod import nim
from multiplug.py_mod import py
from multiplug.rust_mod import rust



#
# (c) 2021-2022 Giorgio Gonnella, University of Goettingen, Germany
#
"""
Programmatically import a Nim module and compile it on the
fly if necessary.
"""

import sh
import sys
import importlib
from pathlib import Path
import nimporter
import os
from contextlib import contextmanager
from multiplug.plugin_api import enforce_plugin_api

NIM_CONST_PFX="py_const_"

def _import_constants(m, prefix):
  for k in list(m.__dict__.keys()):
    if k.startswith(prefix):
      c = k[len(prefix):]
      setattr(m, c, getattr(m, k)())
      delattr(m, k)

def nim(filename, verbose=False, req_const=[], opt_const=[],
        req_func=[],  opt_func=[], const_pfx=NIM_CONST_PFX):
  """
  Import (and compile if necessary) a module written in Nim and
  based on Nimpy.
  """
  modulename = Path(filename).stem
  parent = Path(filename).parent
  m = nimporter.Nimporter.import_nim_module(modulename, [parent])
  info = [f"# nim module {modulename} imported from file {filename}\n"]
  if len(const_pfx) > 0:
    _import_constants(m, const_pfx)
    info += enforce_plugin_api(m, modulename, req_func=req_func,
                             opt_func=opt_func, req_const=req_const,
                             opt_const=opt_const)
  else:
    assert len(req_const) == 0
    info += enforce_plugin_api(m, modulename, req_func=req_func,
                             opt_func=opt_func, opt_const=opt_const)
    info.append("# constants definition mechanism disabled\n")
  if verbose:
    sys.stderr.write("".join(info))
  m.__lang__ = "nim"
  return m




#
# (c) 2022 Giorgio Gonnella, University of Goettingen, Germany
#
PLUGIN_METADATA_SCALAR_KEYS = ["ID", "VERSION", "INPUT",
                               "METHOD", "IMPLEMENTATION", "REQ_SOFTWARE",
                               "REQ_HARDWARE", "ADVICE"]
PLUGIN_METADATA_LIST_KEYS = ["OUTPUT"]
PLUGIN_METADATA_TUPLES_LIST_KEYS = ["PARAMETERS"]

def check_metadata_keys_defined():
  """
  Check that plugin metadata key definitions have been provided.
  """
  if "PLUGIN_METADATA_SCALAR_KEYS" and "PLUGIN_METADATA_LIST_KEYS" and \
      "PLUGIN_METADATA_TUPLES_LIST_KEYS" not in globals():
    raise Exception("Plugin metadata keys not defined.\n" \
                "Please define at least one of the following constants:\n" \
                "PLUGIN_METADATA_SCALAR_KEYS, PLUGIN_METADATA_LIST_KEYS, " \
                "PLUGIN_METADATA_TUPLES_LIST_KEYS\n")

def metadata_str(plugin):
  """
  Returns a dictionary with string representations of the metadata of the
  plugin, based on constants defined in the plugin code.
  """
  check_metadata_keys_defined()
  result_s = {k.lower(): getattr(plugin, k) for \
      k in PLUGIN_METADATA_SCALAR_KEYS if hasattr(plugin, k)}

  result_l = {k.lower(): getattr(plugin, k) for \
      k in PLUGIN_METADATA_LIST_KEYS if hasattr(plugin, k)}
  for k in result_l.keys():
    result_l[k] = ",".join([str(x) for x in result_l[k]])

  result_t = {k.lower(): getattr(plugin, k) for \
      k in PLUGIN_METADATA_TUPLES_LIST_KEYS if hasattr(plugin, k)}
  for k in result_t.keys():
    result_t[k] = ";".join([",".join(str(x)) for x in result_t[k]])

  return {**result_s, **result_l, **result_t}
